# -*- coding: utf-8 -*-

# Note: created with cookiecutter by someone with no experience of how
# to make a pytest plugin. Please question anything related to the
# pytest integration!

import re
import os
import io
import contextlib

import param

import pytest
import nbformat
import nbconvert
from nbconvert.preprocessors import ExecutePreprocessor


__version__ = str(param.version.Version(
    fpath=__file__, archive_commit="$Format:%h$", reponame="panel"))

from .lint import LintNb
from .verify import VerifyNb



def pytest_addoption(parser):
    group = parser.getgroup('nbsmoke')
    group.addoption(
        '--nbsmoke-run',
        action="store_true",
        help="Run notebooks using nbconvert to check for exceptions.")

    group.addoption(
        '--nbsmoke-lint',
        action="store_true",
        help="Lint check notebooks using flake8")

    group.addoption(
        '--nbsmoke-lint-debug',
        action="store_true",
        help="Write out copy of python script resulting from conversion of ipynb")

    group.addoption(
        '--nbsmoke-lint-onlywarn',
        action="store_true",
        help="Flake errors will only appear as warnings")

    group.addoption(
        '--nbsmoke-verify',
        action="store_true",
        help="Verify notebooks")

    group.addoption(
        '--store-html',
        action="store",
        dest='store_html',
        default='',
        help="When running, store rendered-to-html notebooks in the supplied path.")

    parser.addini('nbsmoke_cell_timeout', "nbsmoke's nbconvert cell timeout")

    ####
    # TODO: hacks to work around pyviz team desire to not use pytest's markers
    parser.addini('nbsmoke_skip_run', 're to skip (multi-line; one pattern per line)')
    group.addoption(
        '--ignore-nbsmoke-skip-run',
        action="store_true",
        help="Ignore any skip list in the ini file (allows to run all nbs if desired)")
    ####

    # TODO: remove/rename/see pytest python_files
    parser.addini('it_is_nb_file', 're to determine whether file is notebook')

    parser.addini('nbsmoke_flakes_to_ignore', "flake messages to ignore during nbsmoke's flake checking")

    parser.addini('nbsmoke_flakes_cell_magics_blacklist', "cell magics you don't want to see - i.e. treat as lint.")
    parser.addini('nbsmoke_flakes_line_magics_blacklist', "line magics you don't want to see - i.e. treat as lint")

    parser.addini('nbsmoke_magic_handlers', "path to .py file containing custom magic handlers")


@contextlib.contextmanager
def cwd(d):
    orig = os.getcwd()
    os.chdir(d)
    try:
        yield
    finally:
        os.chdir(orig)



###################################################


class RunNb(pytest.Item):

    def repr_failure(self, excinfo):
        return excinfo.exconly(True)

    def runtest(self):
        self._skip()
        with io.open(self.name,encoding='utf8') as nb:
            notebook = nbformat.read(nb, as_version=4)

            # TODO: which kernel? run in pytest's or use new one (make it option)
            _timeout = self.parent.parent.config.getini('nbsmoke_cell_timeout')
            kwargs = dict(timeout=int(_timeout) if _timeout!='' else 300,
                          allow_errors=False,
                          # or sys.version_info[1] ?
                          kernel_name='python')

            ep = ExecutePreprocessor(**kwargs)
            with cwd(os.path.dirname(self.name)): # jupyter notebook always does this, right?
                ep.preprocess(notebook,{})

            # TODO: clean up this option handling
            if self.parent.parent.config.option.store_html != '':
                he = nbconvert.HTMLExporter()
                # Backwards incompatible change in nbconvert 6 in template file names
                if nbconvert.version_info[0] < 6:
                    he.template_file = 'basic'
                else:
                    he.template_file = 'classic/base.html.j2'
                # could maybe use this for chance of testing the html? but not the aim of this project
                #he.template_file = 'basic'
                html, resources = he.from_notebook_node(notebook)
                with io.open(os.path.join(self.parent.parent.config.option.store_html,os.path.basename(self.name)+'.html'),'w',encoding='utf8') as f:
                    f.write(html)

    def _skip(self):
        _skip_patterns = self.parent.parent.config.getini('nbsmoke_skip_run')
        if not self.parent.parent.config.option.ignore_nbsmoke_skip_run:
            for pattern in _skip_patterns.splitlines():
                if re.match(pattern,self.nodeid.split("::")[0],re.IGNORECASE):
                    pytest.skip()


class IPyNbFile(pytest.File):
    def __init__(self, fspath, parent=None, config=None, session=None, dowhat=RunNb):
        self._dowhat = dowhat
        super(IPyNbFile,self).__init__(fspath, parent=parent, config=None, session=None)

    def collect(self):
        if hasattr(self._dowhat, "from_parent"):
            yield self._dowhat.from_parent(self,name=str(self.fspath))
        else: # older pytest (https://docs.pytest.org/en/stable/deprecations.html#node-construction-changed-to-node-from-parent)
            yield self._dowhat(str(self.fspath), self)


def pytest_collect_file(path, parent):
    opt = parent.config.option
    # TODO: Make this pattern standard/configurable.
    # match .ipynb except .nbval.ipynb
    it_is_nb_file = parent.config.getini('it_is_nb_file')
    if it_is_nb_file == '':
        #"^((?!\.nbval).)*\.ipynb$"
        it_is_nb_file = r"^.*\.ipynb"
    if re.match(it_is_nb_file,path.strpath,re.IGNORECASE):
        if opt.nbsmoke_run or opt.nbsmoke_lint or opt.nbsmoke_verify:
            # TODO express via the options system if you ever figure it out
            # Hmm, should be able to do all - clean up!
            assert (opt.nbsmoke_run ^ opt.nbsmoke_lint) ^ opt.nbsmoke_verify
            if opt.nbsmoke_run:
                dowhat = RunNb
            elif opt.nbsmoke_lint:
                dowhat = LintNb
            elif opt.nbsmoke_verify:
                dowhat = VerifyNb
            if hasattr(IPyNbFile, "from_parent"):
                return IPyNbFile.from_parent(parent, fspath=path, dowhat=dowhat)
            else: # for older pytest (https://docs.pytest.org/en/stable/deprecations.html#node-construction-changed-to-node-from-parent)
                return IPyNbFile(path, parent, dowhat=dowhat)



"""Mainly code copied in from datashader's examples/nb, which itself
was copied straight from bokeh? Would be great to coordinate that a
bit better...

"""

import traceback
import warnings
import io
import nbformat
import nbconvert
import sys
import ast
import re

import pytest

class VerifyNb(pytest.Item):
    def runtest(self):

        ###################################################
        # quick hacks to handle missing optional modules
        _missing = []
        if BeautifulSoup is None:
            _missing.append("  * BeautifulSoup not available - please install beautifulsoup4")
        if requests is None:
            _missing.append("  * requests not available - please install requests")
        if _missing:
            raise Exception("nbsmoke verify failed to execute because of missing modules:\n"+"\n".join(_missing))
        ###################################################

        filename = self.name

        bad_modules = check_modules(filename)
        try:
            bad_links = check_urls(filename, name='a', attribute='href')
        except requests.exceptions.ConnectionError:
            bad_links = traceback.format_exc(1)

        bad_images = check_urls(filename, name='img', attribute='src')

        if bad_modules or bad_links or bad_images:
            warnings.warn("%s:\n***** invalid modules: %s\n\n***** bad links: %s\n\n***** bad images: %s\n\n"%(filename,bad_modules,bad_links,bad_images) + "-"*20)


# TODO: capture or log the import errors
try:
    import requests, requests.exceptions
except:
    requests = None

try:
    from bs4 import BeautifulSoup
except:
    BeautifulSoup = None

###################################################
###################################################
# rest of this file is quickly copied in code from datashader
# examples/nb (which itself came from bokeh). Need to check it works
# as expected. Takes a while to run so must be doing something :) But
# fails to run properly for various projects other than datashader
# (e.g. for geoviews, gives spurious errors about notebook syntax,
# complains that
# https://www.earthsystemcog.org/projects/esmf/regridding is a bad
# link).

def export_as_html(filename):
    html_exporter = nbconvert.HTMLExporter()
    # Backwards incompatible change in nbconvert 6 in template file names
    if nbconvert.version_info[0] < 6:
        html_exporter.template_file = 'basic'
    else:
        html_exporter.template_file = 'classic/base.html.j2'
    body, _ = html_exporter.from_filename(filename)
    return body

def export_as_python(filename):
    with io.open(filename,encoding='utf8') as nbfile:
        nb = nbformat.read(nbfile, as_version=4)
        output, _ = nbconvert.PythonExporter().from_notebook_node(nb)
        if sys.version_info[0]==2:
            # notebooks will start with "coding: utf-8", but output already unicode
            # (alternative hack would be to find the coding line and remove it)
            output = output.encode('utf8')
        return output


def module_exists(name):
    try:
        __import__(name)
    except ImportError:
        return False
    else:
        return True


def url_exists(url):
    headers = {'User-Agent': 'Mozilla/5.0'}  # dummy user agent for security filters
    response = requests.head(url, headers=headers)
    if not response.ok:
        response = requests.get(url, headers=headers)
    return response.ok


def check_modules(notebook):
    class ModuleReader(ast.NodeVisitor):
        def __init__(self):
            self.imports = set()

        def generic_visit(self, node):
            ast.NodeVisitor.generic_visit(self, node)
            return list(self.imports)

        def visit_Import(self, node):
            for alias in node.names:
                self.imports.add(alias.name)

        def visit_ImportFrom(self, node):
            self.imports.add(node.module)

    def get_ipython_modules(s):
        root = ast.parse(s)
        return ModuleReader().visit(root)

    bad_modules = set()

    for module in get_ipython_modules(export_as_python(notebook)):
        m = re.sub(r'\..*', '', module) if '.' in module else module
        if not module_exists(m):
            bad_modules.add(m)

    return bad_modules


def check_urls(notebook, name, attribute):
    REGEX_URL = re.compile('^(http|https):.+')

    bad_urls = set()

    html = export_as_html(notebook)
    soup = BeautifulSoup(html, 'html.parser')
    for tag in soup.find_all(name):
        url = tag.get(attribute)
        if REGEX_URL.match(url) and not url_exists(url):
            bad_urls.add(url)

    return bad_urls




"""
Presumably I copied this in from holoviews and hacked til it worked,
as a proof of concept? But since holoviews is deprecating magics, no
attempt was ever made (or will ever be made...) to do it properly :)
"""

from itertools import groupby

IGNORED_LINE_MAGICS = ['output']
IGNORED_CELL_MAGICS = ['output']

try:
    import numpy # noqa: Some bad numpy/pytest interaction. Without importing numpy here (so it happens when nbsmoke is imported, i.e. at plugin load time,
                 # we get the traceback in https://github.com/numpy/numpy/issues/14012 (no idea if same cause)
except ImportError:
    pass

def _make_optsspec():
    from holoviews.util.parser import OptsSpec
    
    class NbSmokeOptsSpec(OptsSpec):
    
        @classmethod
        def _hvparse(cls, line, ns={}):
            """
            Parse an options specification, returning a dictionary with
            path keys and {'plot':<options>, 'style':<options>} values.
            """
            parses  = [p for p in cls.opts_spec.scanString(line)]
            if len(parses) != 1:
                raise SyntaxError("Invalid specification syntax.")
            else:
                e = parses[0][2]
                processed = line[:e]
                if (processed.strip() != line.strip()):
                    raise SyntaxError("Failed to parse remainder of string: %r" % line[e:])
        
            grouped_paths = cls._group_paths_without_options(cls.opts_spec.parseString(line))
            things = []
            for pathspecs, group in grouped_paths:
        
        #        normalization = cls.process_normalization(group)
        #        if normalization is not None:
        #            options['norm'] = normalization
        
                if 'plot_options' in group:
                    plotopts =  group['plot_options'][0]
                    opts = cls.todict(plotopts, 'brackets', ns=ns)
                    things+=opts
        
                if 'style_options' in group:
                    styleopts = group['style_options'][0]
                    opts = cls.todict(styleopts, 'parens', ns=ns)
                    things+=opts
        
            return things
        
        @classmethod
        def _hvtodict(cls, parseresult, mode='parens', ns={}):
            """
            Helper function to return dictionary given the parse results
            from a pyparsing.nestedExpr object (containing keywords).
        
            The ns is a dynamic namespace (typically the IPython Notebook
            namespace) used to update the class-level namespace.
            """
            grouped = []
            things = []
            tokens = cls.collect_tokens(parseresult, mode)
            # Group tokens without '=' and append to last token containing '='
            for group in groupby(tokens, lambda el: '=' in el):
                (val, items) = group
                if val is True:
                    grouped += list(items)
                if val is False:
                    elements =list(items)
                    # Assume anything before ) or } can be joined with commas
                    # (e.g tuples with spaces in them)
                    joiner=',' if any(((')' in el) or ('}' in el))
                                      for el in elements) else ''
                    grouped[-1] += joiner + joiner.join(elements)
        
            for keyword in grouped:
                # Tuple ('a', 3) becomes (,'a',3) and '(,' is never valid
                # Same for some of the other joining errors corrected here
                for (fst,snd) in [('(,', '('), ('{,', '{'), ('=,','='),
                                  (',:',':'), (':,', ':'), (',,', ','),
                                  (',.', '.')]:
                    keyword = keyword.replace(fst, snd)
        
                things.append('dict(%s)' % keyword)
        
            return things

    return NbSmokeOptsSpec
        

def opts_handler(magic):
    """Given an opts magic, return line of python suitable for pyflakes."""

    NbSmokeOptsSpec = _make_optsspec()
    string_of_magic_args = magic.python
    return " ; ".join(NbSmokeOptsSpec.parse(string_of_magic_args)) + " # noqa: here to use names for original %s magic: %s"%(magic.__class__.__name__, magic.python)


cell_magic_handlers = {'opts': opts_handler}
line_magic_handlers = {'opts': opts_handler}



"""support for ipython magics during notebook flake checking

Call insert_get_ipython(notebook) at the start so pyflakes knows about
ipython, then use ipython_to_python_for_flake_checks() to replace
various forms of magics with python in a form that makes sense for
flake checking. See ipython_to_python_for_flake_checks() for more
info.

"""

import sys
import warnings
import ast

import nbformat


# try to fight the advertising/deprecation warnings and get an
# ipython2python function without confusing messages appearing
# all over the place
try:
    from IPython.core.inputtransformer2 import TransformerManager
    ipython2python = TransformerManager().transform_cell
except:
    # this is the import I wanted to use, but it generates
    # deprecation wanrings when all the latest packages are
    # installed (as of 12 June 2019).
    from nbconvert.filters.strings import ipython2python

from . import holoviews_support
from . import builtins_support

# TODO: still nede to investigate the following line magics (and what
# does ipython convert do to them?):
#
#   ? cd
#   ? dirs
#    macro
#   popd
#   pushd
#   pwd
#   pylab (it imports all numpy, pylab, etc etc)
#   rehashx
#   reset?
#   reset_selective
#   sc
#   sx
#   system
#   tb
#   xdel??
#   xmode

# maybe skip?:
#
#  ? alias
#  ? alias_magic
#  ? autoawait
#  ?

# ???
# autocall
# automagic
#

# TODO: Allow people to ignore stuff per project.  E.g. allow
# to add to SIMPLE_MAGICS, IGNORED_MAGICS,
# other_magic_handlers via command-line arg and ini file. And
# point to that mechanism in this warning.

# these will be ignored (i.e. will not be present in python that gets
# lint checked)
IGNORED_LINE_MAGICS = []
IGNORED_CELL_MAGICS = []

# things you don't want to see in your notebooks
BLACKLISTED_LINE_MAGICS = []
BLACKLISTED_CELL_MAGICS = []

# "%magic xyz..." will be replaced by "xyz..."
SIMPLE_LINE_MAGICS = ['time','timeit','prun']

# fns that take some particular magic and turn it into a python line
# that is appropriate for linting (e.g. could return a no-op that uses
# an imported name, or just "pass", etc).
# TODO: the handlers now have a bad interface and are side effecty
other_line_magic_handlers = {}
other_cell_magic_handlers = {}

for hmm in (builtins_support, holoviews_support):
    other_line_magic_handlers.update(hmm.line_magic_handlers)
    other_cell_magic_handlers.update(hmm.cell_magic_handlers)
    IGNORED_LINE_MAGICS.extend(hmm.IGNORED_LINE_MAGICS)
    IGNORED_CELL_MAGICS.extend(hmm.IGNORED_CELL_MAGICS)

# TODO: suddenly had to make some fns into a class to support blacklists; should rework.
class Processor(object):

    def __init__(self, extra_cell_blacklist=None, extra_line_blacklist=None, extra_magic_handlers=None):
        self.blacklisted_cell = (extra_cell_blacklist or []) + BLACKLISTED_CELL_MAGICS
        self.blacklisted_line = (extra_line_blacklist or []) + BLACKLISTED_LINE_MAGICS
        self.extra_cell_magic_handlers = dict(other_cell_magic_handlers)
        self.extra_line_magic_handlers = dict(other_line_magic_handlers)

        if extra_magic_handlers:
            user_cell_magic_handlers, user_line_magic_handlers = self._load_user_magic_handlers(extra_magic_handlers)
            self.extra_cell_magic_handlers.update(user_cell_magic_handlers)
            self.extra_line_magic_handlers.update(user_line_magic_handlers)


    @staticmethod
    def _load_user_magic_handlers(path):
        from importlib import util
        spec = util.spec_from_file_location("_nbsmoke_user_magic_handlers", path)
        mod = util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        return mod.cell_magic_handlers, mod.line_magic_handlers
    
                                                                      
    @staticmethod
    def insert_get_ipython(nb):
        # define and use get_ipython (for pyflakes)
        if len(nb['cells']) > 0:
            # the get_ipython() is so pyflakes doesn't complain if no
            # magics present (which would leave get_ipython unused)
            get_ipython_cell = nbformat.v4.new_code_cell(
                'from IPython import get_ipython\nget_ipython()')
            nb['cells'].insert(0,get_ipython_cell)


    def ipython_to_python_for_flake_checks(self, ipy):
        """Given some ipython, return python code suitable for flake checking.

        Regular python (non-magic) lines will be left alone.

        Zero-arg magics will be ignored (omitted from generated python).

        Certain other magics will also be ignored (omitted from
        generated python), e.g. "matplotlib" (see IGNORED_LINE_MAGICS,
        IGNORED_CELL_MAGICS).

        Certain single-argument magics that wrap some regular python will
        be unwrapped: (e.g. "%time fn()" will be replaced by "fn()"; see
        SIMPLE_LINE_MAGICS).

        Optional "magic handlers" can be registered to deal with
        custom/third-party magics (see other_line_magic_handlers,
        other_cell_magic_handlers).

        If a magic is encountered that wasn't handled in any of the above
        ways, it'll be omitted, but with a warning.

        Finally, after the above, if a magic appears in the
        user-supplied cell or line magics blacklist, or is in
        BLACKLISTED_CELL_MAGICS or BLACKLISTED_LINE_MAGICS, it will be
        flagged as a flake.
        """
        ipy_lines = ipy.split('\n') # this is valid for notebooks (json) - i.e. can't be other line endings?
        py_lines = []
        self._transform_ipy_to_py(ipy_lines, py_lines)
        return "\n".join(py_lines)

    def _transform_ipy_to_py(self, lines_in, lines_out):
        for line in lines_in:
            magic_parser = _get_parser(line)
            lines_out.append(self._process(magic_parser))
            self._transform_ipy_to_py(magic_parser.additional_lines, lines_out)

    def _process(self, magic):
        if isinstance(magic, NotMagic):
            content = magic.line
        elif isinstance(magic, LineMagic):
            content = _process_magics(magic, self.extra_line_magic_handlers, IGNORED_LINE_MAGICS, self.blacklisted_line, SIMPLE_LINE_MAGICS)
        elif isinstance(magic, CellMagic):
            content = _process_magics(magic, self.extra_cell_magic_handlers, IGNORED_CELL_MAGICS, self.blacklisted_cell)
        else:
            raise
        return content


def _get_parser(line): # yuck
    for start,cls in parsers.items():
        if start is not None and line.lstrip().startswith(start):
            return cls(line)
    return parsers[None](line) # default/no magic

def _call_a_handler(handlers, magic):
    try:
        return handlers[magic.name](magic)
    except ImportError as e:
        raise ImportError("nbsmoke can't process the following '%s' magic without additional dependencies:\n  %s\n. Error was:\n  %s"%(magic.name, magic.line, e))

def _process_magics(magic, other_magic_handlers, ignored_magics, blacklisted_magics, simple_magics=None):
    if simple_magics is None:
        simple_magics = []

    if magic.name in other_magic_handlers:
        content = _call_a_handler(other_magic_handlers, magic)
    elif magic.skippable:
        # pass with original line as comment is just to help with debugging
        content = 'pass # skipped original line: %s'%magic.line
    elif magic.name in ignored_magics:
        content = 'pass # deliberately ignored %s: %s %s'%(magic.__class__.__name__,magic.name,magic.python)
    elif magic.name in simple_magics:
        content = magic.python
    else:
        # TODO: When users can configure simple/ignored etc magics,
        # this warning could point to that mechanism.
        w = "nbsmoke doesn't know how to process the '%s' %s and has ignored it. Line:\n%s\nPlease file an issue at github.com/pyviz/nbsmoke/issues if it should be processed as part of flake checks, or if it should be silently ignored during flake checks."%(magic.name,magic.__class__.__name__,magic.line)
        if sys.version_info[0] == 2:
            # otherwise you get "UnicodeWarning: Warning is using unicode non convertible to ascii, converting to a safe representation" from pytest
            w = w.encode('utf8')
        warnings.warn(w)
        content = 'pass # was: %s'%magic.line

    if magic.name in blacklisted_magics:
        content += ' # nbsmoke-blacklisted: %s'%magic.name

    return magic.indent + content


# the following 'parser' classes could likely be simplified (there
# were more of them, but I just cut a few out).

# abstract
class _Parser(object):
    """For parsing 'a line of python' that was produced by nbconvert ipynb
    -> python.

    'a line of python' could correspond to more than one line
    in the original notebook; cell magics will have been converted to
    something like "fn(magic..., lines)", where lines is the cell content.

    The type of magic (line or cell) is based on the start of the line
    - see "start".

    """
    start = None
    def __init__(self,line):
        """Attributes:

          * line: the original line

          * indent: of the original line

          * name: the name of the magic (could be None).

          * python: some valid bit of python taken from the original
            line, suitable for taking part in lint check.

          * skippable: whether the python can be skipped from linting
            (because it's empty).

          * additional_lines: some magics (e.g. Cell) can contain
            other lines that will themselves need processing.

        """
        self.line = line
        self.indent = self._get_indent(line)
        pre_line = self._pre(line)
        self.additional_lines = self._get_additional_lines(pre_line)
        self.name = self._get_name(pre_line)
        self.python = self._get_python(pre_line)
        self.skippable = (self.python.strip() == '') # e.g. zero args

    @staticmethod
    def _pre(line):
        return line.strip()

    @staticmethod
    def _get_indent(line):
        return " " * (len(line) - len(line.lstrip()))

    @classmethod
    def _get_name(cls, pre_line):
        return None

    @classmethod
    def _get_python(cls, pre_line):
        return pre_line

    def _get_additional_lines(self, pre_line):
        return []

class NotMagic(_Parser):
    """Line wasn't a magic"""
    pass

# abstract
class _Magic(_Parser):

    @staticmethod
    def _parse(x,i):
        return ast.parse(x).body[0].value.args[i].s

    @classmethod
    def _get_name(cls, pre_line):
        return cls._parse(pre_line, 0)

    @classmethod
    def _get_python(cls, pre_line):
        return cls._parse(pre_line, 1)

class CellMagic(_Magic):
    """cell magic that was converted to fn by nbconvert"""
    start = 'get_ipython().run_cell_magic('

    def _get_additional_lines(self, pre_line):
        remaining_lines = self._parse(pre_line, 2).splitlines()
        return ipython2python("\n".join(remaining_lines)).splitlines()

class LineMagic(_Magic):
    """line magic that was converted to fn by nbconvert"""
    start = 'get_ipython().run_line_magic('

class Py2LineMagic(LineMagic):
    """line magic that was converted to fn by nbconvert under python 2"""
    start = 'get_ipython().magic('

    @staticmethod
    def _parse(x,i):
        # TODO (py2): should find and use ipython's own code for
        # parsing instead (but py2 is going away...)
        bits = LineMagic._parse(x, 0).split(" ", 1)
        if i == 0:
            return bits[i]
        else:
            return bits[i] if len(bits)>i else ''

parsers = {}
for cls in (NotMagic, CellMagic, LineMagic, Py2LineMagic):
    assert cls.start not in parsers
    parsers[cls.start] = cls
